#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <richedit.h>
#include "AkelDLL.h"


//Functions prototypes
LRESULT CALLBACK NewEditParentProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
void SaveClipboard(UINT uFormat, wchar_t **wpData, char **pData);
void SetClipboardFormat(UINT uFormat);
void InitCommon(PLUGINDATA *pd);
void InitMain();
void UninitMain();

//Global variables
HWND hMainWnd;
BOOL bOldWindows;
BOOL bMDI;
BOOL bInitCommon=FALSE;
BOOL bInitMain=FALSE;
BOOL bSelAutoCopyActive=FALSE;
BOOL bSelAutoFirstRun=FALSE;
UINT cfSelAutoCopy=0;
char *pClipboard=NULL;
wchar_t *wpClipboard=NULL;
WNDPROCDATA *NewMainProcData=NULL;
WNDPROCDATA *NewFrameProcData=NULL;


//Identification
extern "C" void __declspec(dllexport) DllAkelPadID(PLUGINVERSION *pv)
{
  pv->dwAkelDllVersion=AKELDLL;
  pv->dwExeMinVersion3x=MAKE_IDENTIFIER(3, 7, 4, 0);
  pv->dwExeMinVersion4x=MAKE_IDENTIFIER(4, 4, 0, 0);
  pv->pPluginName="SelAutoCopy";
}

//Plugin extern function
extern "C" void __declspec(dllexport) Main(PLUGINDATA *pd)
{
  if (pd->lpbAutoLoad)
  {
    *pd->lpbAutoLoad=TRUE;
    return;
  }

  if (!bInitCommon) InitCommon(pd);

  //Is plugin already loaded?
  if (bInitMain)
  {
    bSelAutoCopyActive=FALSE;
    UninitMain();
  }
  else
  {
    InitMain();

    if (!pd->bOnStart)
    {
      SELCHANGE sc;

      bSelAutoCopyActive=TRUE;
      bSelAutoFirstRun=TRUE;

      sc.nmhdr.hwndFrom=pd->hWndEdit;
      sc.nmhdr.idFrom=ID_EDIT;
      sc.nmhdr.code=EN_SELCHANGE;
      SendMessage(pd->hWndEdit, EM_EXGETSEL, 0, (LPARAM)&sc.chrg);
      NewEditParentProc(GetParent(pd->hWndEdit), WM_NOTIFY, ID_EDIT, (LPARAM)&sc);
    }

    //Stay in memory, and show as active
    pd->nUnload=UD_NONUNLOAD_ACTIVE;
  }
}

extern "C" void __declspec(dllexport) Paste(PLUGINDATA *pd)
{
  if (pd->lpbAutoLoad)
  {
    *pd->lpbAutoLoad=FALSE;
    return;
  }

  if (!bInitCommon) InitCommon(pd);

  if (bInitMain)
  {
    if (IsClipboardFormatAvailable(cfSelAutoCopy))
    {
      if (wpClipboard)
      {
        SendMessage(pd->hMainWnd, AKD_REPLACESELW, (WPARAM)pd->hWndEdit, (LPARAM)wpClipboard);
      }
      else if (pClipboard)
      {
        SendMessage(pd->hMainWnd, AKD_REPLACESELA, (WPARAM)pd->hWndEdit, (LPARAM)pClipboard);
      }
    }
    else SendMessage(pd->hMainWnd, AKD_PASTE, (WPARAM)pd->hWndEdit, FALSE);

    //Stay in memory, but show as non-active
    pd->nUnload=UD_NONUNLOAD_NONACTIVE;
  }
}

LRESULT CALLBACK NewEditParentProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  if (uMsg == AKDN_MAIN_ONSTART_FINISH)
  {
    bSelAutoCopyActive=TRUE;
  }
  else if (uMsg == WM_NOTIFY)
  {
    if (bSelAutoCopyActive)
    {
      if (wParam == ID_EDIT)
      {
        if (((NMHDR *)lParam)->code == EN_SELCHANGE)
        {
          if (((SELCHANGE *)lParam)->chrg.cpMin != ((SELCHANGE *)lParam)->chrg.cpMax)
          {
            if (((SELCHANGE *)lParam)->chrg.cpMax - ((SELCHANGE *)lParam)->chrg.cpMin == 1)
            {
              TEXTRANGEA txtrng;
              char szChar[3];

              txtrng.chrg.cpMin=((SELCHANGE *)lParam)->chrg.cpMin;
              txtrng.chrg.cpMax=((SELCHANGE *)lParam)->chrg.cpMax;
              txtrng.lpstrText=szChar;

              if (SendMessageA(hWnd, EM_GETTEXTRANGE, 0, (LPARAM)&txtrng))
              {
                if (szChar[0] != '\0')
                {
                  SaveClipboard(cfSelAutoCopy, &wpClipboard, &pClipboard);
                  SendMessage(hMainWnd, AKD_COPY, (WPARAM)((NMHDR *)lParam)->hwndFrom, 0);
                  SetClipboardFormat(cfSelAutoCopy);
                }
              }
            }
            else
            {
              SaveClipboard(cfSelAutoCopy, &wpClipboard, &pClipboard);
              SendMessage(hMainWnd, AKD_COPY, (WPARAM)((NMHDR *)lParam)->hwndFrom, 0);
              SetClipboardFormat(cfSelAutoCopy);
            }
          }
        }
      }
      if (bSelAutoFirstRun)
      {
        bSelAutoFirstRun=FALSE;
        return 0;
      }
    }
  }

  //Call next procedure
  if (hWnd == hMainWnd)
    return NewMainProcData->NextProc(hWnd, uMsg, wParam, lParam);
  else
    return NewFrameProcData->NextProc(hWnd, uMsg, wParam, lParam);
}

void SaveClipboard(UINT uFormat, wchar_t **wpData, char **pData)
{
  HGLOBAL hDataSource;
  LPVOID pDataSource;
  int nLen;

  if (OpenClipboard(NULL))
  {
    if (!IsClipboardFormatAvailable(uFormat))
    {
      if (hDataSource=GetClipboardData(CF_UNICODETEXT))
      {
        if (pDataSource=GlobalLock(hDataSource))
        {
          nLen=wcslen((wchar_t *)pDataSource) * sizeof(wchar_t) + 2;

          if (*wpData)
          {
            GlobalFree((HGLOBAL)*wpData);
            *wpData=NULL;
          }
          if (*pData)
          {
            GlobalFree((HGLOBAL)*pData);
            *pData=NULL;
          }

          if (*wpData=(wchar_t *)GlobalAlloc(GMEM_FIXED, nLen))
          {
            memcpy(*wpData, pDataSource, nLen);
          }
          GlobalUnlock(hDataSource);
        }
      }
      else if (hDataSource=GetClipboardData(CF_TEXT))
      {
        if (pDataSource=GlobalLock(hDataSource))
        {
          nLen=lstrlenA((char *)pDataSource) + 1;

          if (*wpData)
          {
            GlobalFree((HGLOBAL)*wpData);
            *wpData=NULL;
          }
          if (*pData)
          {
            GlobalFree((HGLOBAL)*pData);
            *pData=NULL;
          }

          if (*pData=(char *)GlobalAlloc(GMEM_FIXED, nLen))
          {
            memcpy(*pData, pDataSource, nLen);
          }
          GlobalUnlock(hDataSource);
        }
      }
    }
    CloseClipboard();
  }
}

void SetClipboardFormat(UINT uFormat)
{
  if (!IsClipboardFormatAvailable(uFormat))
  {
    if (OpenClipboard(NULL))
    {
      SetClipboardData(uFormat, NULL);
      CloseClipboard();
    }
  }
}

void InitCommon(PLUGINDATA *pd)
{
  bInitCommon=TRUE;
  hMainWnd=pd->hMainWnd;
  bMDI=pd->bMDI;
  bOldWindows=pd->bOldWindows;
}

void InitMain()
{
  bInitMain=TRUE;

  cfSelAutoCopy=RegisterClipboardFormatA("AkelPad::SelAutoCopy");

  //SubClass
  NewMainProcData=NULL;
  SendMessage(hMainWnd, AKD_SETMAINPROC, (WPARAM)NewEditParentProc, (LPARAM)&NewMainProcData);

  if (bMDI)
  {
    NewFrameProcData=NULL;
    SendMessage(hMainWnd, AKD_SETFRAMEPROC, (WPARAM)NewEditParentProc, (LPARAM)&NewFrameProcData);
  }
}

void UninitMain()
{
  bInitMain=FALSE;

  //Remove subclass
  if (NewMainProcData)
  {
    SendMessage(hMainWnd, AKD_SETMAINPROC, (WPARAM)NULL, (LPARAM)&NewMainProcData);
    NewMainProcData=NULL;
  }
  if (bMDI)
  {
    if (NewFrameProcData)
    {
      SendMessage(hMainWnd, AKD_SETFRAMEPROC, (WPARAM)NULL, (LPARAM)&NewFrameProcData);
      NewFrameProcData=NULL;
    }
  }

  //Free memory
  if (wpClipboard)
  {
    GlobalFree((HGLOBAL)wpClipboard);
    wpClipboard=NULL;
  }
  if (pClipboard)
  {
    GlobalFree((HGLOBAL)pClipboard);
    pClipboard=NULL;
  }
}

//Entry point
extern "C" BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
  if (fdwReason == DLL_PROCESS_ATTACH)
  {
  }
  else if (fdwReason == DLL_THREAD_ATTACH)
  {
  }
  else if (fdwReason == DLL_THREAD_DETACH)
  {
  }
  else if (fdwReason == DLL_PROCESS_DETACH)
  {
    if (bInitMain) UninitMain();
  }
  return TRUE;
}
